// SPDX-FileCopyrightText: 2024 Petros Koutsolampros
//
// SPDX-License-Identifier: GPL-3.0-only

#include "salalib/pointmap.h"
#include "salalib/vgamodules/vgametric.h"
#include "salalib/vgamodules/vgavisualglobal.h"
#include "salalib/vgamodules/vgavisuallocal.h"
#include "salalib/vgamodules/vgathroughvision.h"
#include "salalib/vgamodules/vgaangular.h"
#include "salalib/vgamodules/vgaisovist.h"

#include "communicator.h"

#include <Rcpp.h>

// [[Rcpp::export("Rcpp_VGA_throughVision")]]
Rcpp::List vgaThroughVision(Rcpp::XPtr<PointMap> pointMapPtr,
                            const Rcpp::Nullable<bool> copyMapNV = R_NilValue) {
  bool copyMap = true;
  if (copyMapNV.isNotNull()) {
    copyMap = Rcpp::as<bool>(copyMapNV);
  }
  if (copyMap) {
    auto prevPointMap = pointMapPtr;
    const auto &prevRegion = prevPointMap->getRegion();
    pointMapPtr = Rcpp::XPtr(new PointMap(prevRegion));
    pointMapPtr->copy(*prevPointMap, true, true);
  }
  Rcpp::List result = Rcpp::List::create(
    Rcpp::Named("completed") = false
  );
  auto analysisResult = VGAThroughVision().run(
    getCommunicator(true).get(),
    *pointMapPtr,
    false);
  result["completed"] = analysisResult.completed;
  result["newAttributes"] = analysisResult.getAttributes();
  result["mapPtr"] = pointMapPtr;
  return result;
}

// [[Rcpp::export("Rcpp_VGA_angular")]]
Rcpp::List vgaAngular(Rcpp::XPtr<PointMap> pointMapPtr,
                      double radius,
                      bool gatesOnly,
                      const Rcpp::Nullable<bool> copyMapNV = R_NilValue) {

  bool copyMap = true;
  if (copyMapNV.isNotNull()) {
    copyMap = Rcpp::as<bool>(copyMapNV);
  }
  if (copyMap) {
    auto prevPointMap = pointMapPtr;
    const auto &prevRegion = prevPointMap->getRegion();
    pointMapPtr = Rcpp::XPtr(new PointMap(prevRegion));
    pointMapPtr->copy(*prevPointMap, true, true);
  }
  Rcpp::List result = Rcpp::List::create(
    Rcpp::Named("completed") = false
  );
  auto analysisResult = VGAAngular(radius, gatesOnly)
    .run(getCommunicator(true).get(), *pointMapPtr, false);
  result["completed"] = analysisResult.completed;
  result["newAttributes"] = analysisResult.getAttributes();
  result["mapPtr"] = pointMapPtr;
  return result;
}

// [[Rcpp::export("Rcpp_VGA_metric")]]
Rcpp::List vgaMetric(Rcpp::XPtr<PointMap> pointMapPtr,
                     double radius,
                     bool gatesOnly,
                     const Rcpp::Nullable<bool> copyMapNV = R_NilValue) {

  if (radius != -1.0 && radius <= 0) {
    Rcpp::stop("Radius for metric vga must be n (-1) for the whole range or a "
                 "positive number. Got %d", radius);
  }
  bool copyMap = true;
  if (copyMapNV.isNotNull()) {
    copyMap = Rcpp::as<bool>(copyMapNV);
  }
  if (copyMap) {
    auto prevPointMap = pointMapPtr;
    const auto &prevRegion = prevPointMap->getRegion();
    pointMapPtr = Rcpp::XPtr(new PointMap(prevRegion));
    pointMapPtr->copy(*prevPointMap, true, true);
  }
  Rcpp::List result = Rcpp::List::create(
    Rcpp::Named("completed") = false
  );
  auto analysisResult = VGAMetric(radius, gatesOnly)
    .run(getCommunicator(true).get(), *pointMapPtr, false);
  result["completed"] = analysisResult.completed;
  result["newAttributes"] = analysisResult.getAttributes();
  result["mapPtr"] = pointMapPtr;
  return result;
}

// [[Rcpp::export("Rcpp_VGA_visualGlobal")]]
Rcpp::List vgaVisualGlobal(Rcpp::XPtr<PointMap> pointMapPtr,
                           int radius,
                           bool gatesOnly,
                           const Rcpp::Nullable<bool> copyMapNV = R_NilValue) {

  if (radius != -1 && (radius < 1 || radius > 99)) {
    Rcpp::stop("Radius for visibility analysis must be n (-1) for the whole "
                 "range or an integer between 1 and 99 inclusive. Got %i",
                 radius);
  }

  bool copyMap = true;
  if (copyMapNV.isNotNull()) {
    copyMap = Rcpp::as<bool>(copyMapNV);
  }
  if (copyMap) {
    auto prevPointMap = pointMapPtr;
    const auto &prevRegion = prevPointMap->getRegion();
    pointMapPtr = Rcpp::XPtr(new PointMap(prevRegion));
    pointMapPtr->copy(*prevPointMap, true, true);
  }

  Rcpp::List result = Rcpp::List::create(
    Rcpp::Named("completed") = false
  );
  auto analysisResult = VGAVisualGlobal(radius, gatesOnly)
    .run(getCommunicator(true).get(), *pointMapPtr, false);
  result["completed"] = analysisResult.completed;
  result["newAttributes"] = analysisResult.getAttributes();
  result["mapPtr"] = pointMapPtr;
  return result;
}

// [[Rcpp::export("Rcpp_VGA_visualLocal")]]
Rcpp::List vgaVisualLocal(Rcpp::XPtr<PointMap> pointMapPtr,
                          bool gatesOnly,
                          const Rcpp::Nullable<bool> copyMapNV = R_NilValue) {
  bool copyMap = true;
  if (copyMapNV.isNotNull()) {
    copyMap = Rcpp::as<bool>(copyMapNV);
  }
  if (copyMap) {
    auto prevPointMap = pointMapPtr;
    const auto &prevRegion = prevPointMap->getRegion();
    pointMapPtr = Rcpp::XPtr(new PointMap(prevRegion));
    pointMapPtr->copy(*prevPointMap, true, true);
  }

  Rcpp::List result = Rcpp::List::create(
    Rcpp::Named("completed") = false
  );
  auto analysisResult = VGAVisualLocal(gatesOnly)
    .run(getCommunicator(true).get(), *pointMapPtr, false);
  result["completed"] = analysisResult.completed;
  result["newAttributes"] = analysisResult.getAttributes();
  result["mapPtr"] = pointMapPtr;
  return result;
}

// [[Rcpp::export("Rcpp_VGA_isovist")]]
Rcpp::List vgaIsovist(Rcpp::XPtr<PointMap> pointMapPtr,
                      Rcpp::XPtr<ShapeMap> shapeMapPtr,
                      const Rcpp::Nullable<bool> copyMapNV = R_NilValue) {
  bool copyMap = true;
  if (copyMapNV.isNotNull()) {
    copyMap = Rcpp::as<bool>(copyMapNV);
  }
  if (copyMap) {
    auto prevPointMap = pointMapPtr;
    const auto &prevRegion = prevPointMap->getRegion();
    pointMapPtr = Rcpp::XPtr(new PointMap(prevRegion));
    pointMapPtr->copy(*prevPointMap, true, true);
  }

  Rcpp::List result = Rcpp::List::create(
    Rcpp::Named("completed") = false
  );
  auto shapeMap = shapeMapPtr->getAllShapes();

  std::vector<SalaShape> shapes;
  shapes.reserve(shapeMap.size());
  for(auto it = shapeMap.begin(); it != shapeMap.end(); ++it) {
    shapes.push_back(it->second);
  }

  auto analysisResult = VGAIsovist(shapes)
    .run(getCommunicator(true).get(), *pointMapPtr, false);
  result["completed"] = analysisResult.completed;
  result["newAttributes"] = analysisResult.getAttributes();
  result["mapPtr"] = pointMapPtr;
  return result;
}

